# pylint: disable=too-few-public-methods
from http.client import HTTPResponse as Response
from typing import Dict, Optional
from urllib import request
from urllib.parse import parse_qsl, urlencode, urlparse, urlunparse

from easy_notifyer.utils import MultiPartForm, run_in_threadpool


class RequestsBase:
    """Base requests obj"""

    @staticmethod
    def _add_params(url: str, params: Dict) -> str:
        url_parts = list(urlparse(url))
        query = dict(parse_qsl(url_parts[4]))
        query.update(params)
        url_parts[4] = urlencode(query)
        return urlunparse(url_parts)


class Requests(RequestsBase):
    """Client for requests"""

    def __init__(self):
        self._client = request.Request

    def post(
        self,
        *,
        url: str,
        headers: Optional[Dict] = None,
        params: Optional[Dict] = None,
        body: Optional[Dict] = None,
        files: Optional[Dict] = None,
    ) -> Response:
        """Send post request"""
        data = None
        headers = headers or {}
        if params is not None:
            url = self._add_params(url, params)
        if body is not None or files is not None:
            form = MultiPartForm(body=body, files=files)
            data = bytes(form)
            headers = {**headers, **form.header}

        req = self._client(url=url, headers=headers, data=data)
        resp = request.urlopen(req)
        return resp


class AsyncRequests:
    """Async client for requests"""

    def __init__(self):
        self._client = Requests()

    async def post(
        self,
        *,
        url: str,
        params: Optional[Dict] = None,
        headers: Optional[Dict] = None,
        body: Optional[Dict] = None,
        files: Optional[Dict] = None,
    ) -> Response:
        """Send async post request"""
        return await run_in_threadpool(
            self._client.post,
            url=url,
            headers=headers,
            params=params,
            body=body,
            files=files,
        )



import asyncio
import functools
import logging
import traceback
from typing import Callable, List, Optional, Tuple, Type, Union

from easy_notifyer.clients.telegram import Telegram, TelegramAsync
from easy_notifyer.report import Report
from easy_notifyer.utils import generate_filename, run_in_threadpool


logger = logging.getLogger(__name__)


def _report_maker(
    *,
    tback: str,
    func_name: Optional[str],
    header: Optional[str],
    as_attached: bool,
    service_name: Optional[str],
    datetime_format: str,
) -> Report:
    """Make report from.

    Args:
        tback (str): traceback for report.
        func_name (str, optional): name of function when raised error.
        header (str, optional): first line in report message. Default -
        "Your program has crashed ☠️"
        as_attached (bool, optional): make report for sending as a file. Default - False.
        service_name (optional): Service name.
        datetime_format (str, optional): format datetime for report.

    Returns:
        isinstance of Report obj.
    """
    return Report(tback, func_name, header, as_attached, service_name, datetime_format)


def _report_telegram_handler(
    *,
    report: Report,
    token: str,
    chat_id: Union[int, List[int]],
    api_url: Optional[str],
    filename: Optional[str],
    disable_notification: bool,
    disable_web_page_preview: bool,
):
    """Send report.

    Args:
        report (Report): instance of ready to send report.
        token (str): Telegram bot token.
        chat_id (int, list): Chat ids for send message.
        filename (str, optional): make report for sending as a file.
        disable_notification (bool): True to disable notification of message.
        disable_web_page_preview (bool): True to disable web preview for links. Not worked for
            as_attached report.
    """
    bot = Telegram(token=token, chat_id=chat_id, api_url=api_url)
    if report.attach is not None:
        filename = filename or generate_filename()
        bot.send_attach(
            msg=report.report,
            attach=report.attach,
            filename=filename,
            disable_notification=disable_notification,
        )
    else:
        bot.send_message(
            report.report,
            disable_notification=disable_notification,
            disable_web_page_preview=disable_web_page_preview,
        )


async def _async_report_telegram_handler(
    *,
    report: Report,
    token: str,
    chat_id: Union[int, List[int]],
    api_url: Optional[str],
    filename: Optional[str],
    disable_notification: bool,
    disable_web_page_preview: bool,
):
    """Send report.

    Args:
        report (Report): instance of ready to send report.
        token (str): Telegram bot token.
        chat_id (int, list): Chat ids for send message.
        filename (str, optional): make report for sending as a file.
        disable_notification (bool): True to disable notification of message.
        disable_web_page_preview (bool): True to disable web preview for links. Not worked for
            as_attached report.
    """
    bot = TelegramAsync(token=token, chat_id=chat_id, api_url=api_url)
    if report.attach is not None:
        filename = filename or generate_filename()
        await bot.send_attach(
            msg=report.report,
            attach=report.attach,
            filename=filename,
            disable_notification=disable_notification,
        )
    else:
        await bot.send_message(
            report.report,
            disable_notification=disable_notification,
            disable_web_page_preview=disable_web_page_preview,
        )


def telegram_reporter(
    *,
    token: str,
    chat_id: Union[List[int], List[str], int, str],
    api_url: Optional[str] = None,
    service_name: Optional[str] = None,
) -> Union[Callable]:
    """Handler errors for sending report in telegram.

    Args:
        token (str): Telegram bot token. Can be use from environment variable
        To receive: https://core.telegram.org/bots#6-botfather.
        chat_id (int, str, list): Chat ids for send message.
        api_url (str): Url api for telegram.
        service_name (optional): Service name.
    """

    def telegram_wrapper(
        exceptions: Optional[Union[Type[BaseException], Tuple[Type[BaseException], ...]]] = None,
        header: Optional[str] = None,
        as_attached: bool = False,
        datetime_format: str = "%Y-%m-%d %H:%M:%S",
        filename: Optional[str] = None,
        disable_notification: bool = False,
        disable_web_page_preview: bool = False,
    ):
        """Handler errors for sending report in telegram.
        Args:
            exceptions (exception, tuple(exception), optional): Exceptions for handle.
                Two and more - in tuple. Default - Exception.
            header (str, optional): first line in report message.
                Default - "Your program has crashed ☠️"
            as_attached (bool, optional): make report for sending as a file. Default - False.
            datetime_format (str, optional): format datetime for report.
            filename (str, optional): filename for sending report as file.
                Default: datetime %Y-%m-%d %H_%M_%S.txt.
            disable_notification (bool): True to disable notification of message.
            disable_web_page_preview (bool): True to disable web preview for links.
                Not worked for as_attached report.
        """
        if as_attached is True and disable_web_page_preview is True:
            logger.error(
                "Argument disable_web_page_preview not may be True if report mode as_attached"
            )
            disable_web_page_preview = False

        exceptions = exceptions or Exception
        datetime_format = datetime_format

        def decorator(func):
            func_name = func.__name__

            def sync_wrapped_view(*args, **kwargs):
                try:
                    return func(*args, **kwargs)
                except exceptions as exc:
                    tback = traceback.format_exc()
                    report = _report_maker(
                        tback=tback,
                        func_name=func_name,
                        header=header,
                        as_attached=as_attached,
                        service_name=service_name,
                        datetime_format=datetime_format,
                    )
                    _report_telegram_handler(
                        report=report,
                        token=token,
                        chat_id=chat_id,
                        api_url=api_url,
                        filename=filename,
                        disable_notification=disable_notification,
                        disable_web_page_preview=disable_web_page_preview,
                    )
                    raise exc

            async def async_wrapped_view(*args, **kwargs):
                try:
                    return await func(*args, **kwargs)
                except exceptions as exc:
                    tback = traceback.format_exc()
                    report = await run_in_threadpool(
                        _report_maker,
                        tback=tback,
                        func_name=func_name,
                        header=header,
                        as_attached=as_attached,
                        service_name=service_name,
                        datetime_format=datetime_format,
                    )
                    await _async_report_telegram_handler(
                        report=report,
                        token=token,
                        chat_id=chat_id,
                        api_url=api_url,
                        filename=filename,
                        disable_notification=disable_notification,
                        disable_web_page_preview=disable_web_page_preview,
                    )
                    raise exc

            if asyncio.iscoroutinefunction(func):
                return functools.wraps(func)(async_wrapped_view)
            return functools.wraps(func)(sync_wrapped_view)

        return decorator

    return telegram_wrapper



import uuid
from email.mime.application import MIMEApplication
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from smtplib import SMTP, SMTP_SSL
from typing import BinaryIO, List, Optional, Union


class Mailer:
    """Object for send mail"""

    def __init__(
        self,
        *,
        host: str,
        port: int,
        login: str,
        password: str,
        ssl: bool = False,
    ):
        """
        Args:
            host (str, optional): = post of smtp server.
            port (int, optional): = port of smtp server.
            login (str, optional): = login for auth in smtp server.
            password (str, optional): password for auth in smtp server.
            ssl (bool, optional): use SSL connection for smtp.
        """
        self._host = host
        self._port = port
        self._login = login
        self._password = password
        self._ssl = ssl
        self._connection: Optional[SMTP_SSL, SMTP] = None

    def __enter__(self):
        self.connect()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.disconnect()

    def connect(self):
        """Connect to smtp-server and create session"""
        if self._connection is None:
            type_conn = SMTP_SSL if self._ssl is True else SMTP
            self._connection = type_conn(host=self._host, port=self._port)
        self.login()

    def login(self):
        """Create session with login/password"""
        if self._connection is not None and self._login is not None and self._password is not None:
            self._connection.login(user=self._login, password=self._password)

    def disconnect(self):
        """Terminate session"""
        if self._connection is not None:
            self._connection.quit()

    @staticmethod
    def _format_message(
        *,
        from_addr: str,
        to_addrs: List[str],
        text: str,
        subject: str,
        attach: Union[bytes, str, BinaryIO],
        filename: str,
    ) -> MIMEMultipart:
        """Formatting message for send.

        Args:
            from_addr (str): the address sending this mail.
            to_addrs (list(str)): addresses to send this mail to.
            subject (str, optional): subject of the mail.
            attach (bytes, str, tuple, optional): file to send.
            filename (str, optional): filename for attached file.

        Returns:
            MIMEMultipart: multipart of message with body, from, to, attach and subject.
        """
        message = MIMEMultipart()
        message["From"] = from_addr
        message["To"] = ", ".join(to_addrs)
        message["Subject"] = subject
        message.attach(MIMEText(text))

        if attach is not None:
            filename = filename or uuid.uuid4().hex
            if hasattr(attach, "read") and isinstance(attach.read(0), bytes):
                attach = attach.read()
            elif hasattr(attach, "encode"):
                attach = attach.encode()
            message.attach(MIMEApplication(attach, name=filename))
        return message

    def send_message(
        self,
        *,
        message: Optional[str] = None,
        from_addr: str,
        to_addrs: Union[str, List[str]],
        subject: Optional[str] = None,
        attach: Optional[Union[bytes, str, BinaryIO]] = None,
        filename: Optional[str] = None,
    ):
        """Send email.

        Args:
            message (str, optional): Text body of message.
            from_addr (str, optional): the address sending this mail.
            to_addrs (str, list(str), optional): addresses to send this mail to.
            subject (str, optional): subject of the mail.
            attach (bytes, str, tuple, optional): file to send.
            filename (str, optional): filename for attached file.
        """
        to_addrs = [mail.strip() for mail in to_addrs.split(",")]
        msg = self._format_message(
            from_addr=from_addr,
            to_addrs=to_addrs,
            text=message,
            subject=subject,
            attach=attach,
            filename=filename,
        )

        self._connection.sendmail(from_addr=from_addr, to_addrs=to_addrs, msg=msg.as_string())



from easy_notifyer.clients.mailer import Mailer
from easy_notifyer.clients.telegram import Telegram, TelegramAsync
from easy_notifyer.handlers.mailer import mailer_reporter
from easy_notifyer.handlers.telegram import telegram_reporter


__version__ = "0.2.1"


__all__ = [
    "Mailer",
    "Telegram",
    "TelegramAsync",
    "mailer_reporter",
    "telegram_reporter",
]



import asyncio
import functools
from datetime import datetime
from typing import Any, Callable, Dict, Optional
from uuid import uuid4


try:
    import contextvars  # noqa 401
except ImportError:
    contextvars = None


FILENAME_DT_FORMAT = "%Y-%m-%d %H_%M_%S"


class MultiPartForm:
    """Creating body of request"""

    def __init__(self, body: Optional[Dict] = None, files: Optional[Dict] = None):
        self._body = body
        self._files = files
        self.boundary = uuid4().hex

    def __bytes__(self):
        return self.encode()

    @staticmethod
    def _form_data(name: str) -> str:
        """Get formdata for input in body"""
        return f'Content-Disposition: form-data; name="{name}"'

    @staticmethod
    def _attached_file(name: str, filename: str) -> str:
        """Get formdata of file for input in body"""
        return f'Content-Disposition: form-data; name="{name}"; filename="{filename}"'

    @property
    def header(self) -> Dict[str, str]:
        """Get a header request"""
        return {"Content-type": f"multipart/form-data; boundary={self.boundary}"}

    def encode(self) -> bytes:
        """Create a body request."""
        body = []
        boundary = b"--" + self.boundary.encode("utf-8")
        if self._body is not None and isinstance(self._body, dict):
            for name, value in self._body.items():
                body.append(boundary)
                body.append(self._form_data(name).encode("utf-8"))
                body.append(b"")
                body.append(str(value).encode("utf-8"))

        if self._files is not None and isinstance(self._files, dict):
            for fieldname in self._files:
                if len(self._files[fieldname]) != 2:
                    continue
                filename, data = self._files[fieldname]
                body.append(boundary)
                body.append(self._attached_file(fieldname, filename).encode("utf-8"))
                body.append(b"")
                if hasattr(data, "read") is True:
                    data.seek(0)
                    body.append(data.read())
                elif hasattr(data, "encode") is True:
                    body.append(data.encode("utf-8"))

        body.append(boundary + b"--")
        return b"\r\n".join(body)


def generate_filename(date_fmt: Optional[str] = None) -> str:
    """
    Generate of filename for sending report as a file.
    Args:
        date_fmt (str, optional): Format of datetime.
    Returns:
        string of filename.
    """
    date_fmt = date_fmt or FILENAME_DT_FORMAT
    return f"{datetime.now().replace(microsecond=0).strftime(date_fmt)}.txt"


async def run_in_threadpool(func: Callable, *args: Any, **kwargs: Any) -> Any:
    """
    Run sync func in async code in thread pool.
    Args:
        func(callable): func to run
        *args:
        **kwargs:
    Returns:
        same as func
    """
    loop = asyncio.get_event_loop()
    if contextvars is not None:
        child = functools.partial(func, *args, **kwargs)
        context = contextvars.copy_context()
        func = context.run
        args = (child,)
    elif kwargs:
        func = functools.partial(func, **kwargs)
    return await loop.run_in_executor(None, func, *args)



import asyncio
import functools
import traceback
from typing import Callable, List, Optional, Tuple, Type, Union

from easy_notifyer.clients.mailer import Mailer
from easy_notifyer.report import Report
from easy_notifyer.utils import generate_filename, run_in_threadpool


def _report_mailer_handler(
    *,
    report: Report,
    host: str,
    port: int,
    login: str,
    password: str,
    from_addr: str,
    to_addrs: Union[str, List[str]],
    ssl: bool = False,
    filename: Optional[str] = None,
    subject: Optional[str] = None,
):
    with Mailer(
        host=host,
        port=port,
        login=login,
        password=password,
        ssl=ssl,
    ) as mailer:
        mailer.send_message(
            message=report.report,
            from_addr=from_addr,
            to_addrs=to_addrs,
            subject=subject,
            attach=report.attach,
            filename=filename or generate_filename(),
        )


def mailer_reporter(
    *,
    host: str,
    port: int,
    login: str,
    password: str,
    from_addr: str,
    to_addrs: Union[str, List[str]],
    ssl: bool = False,
    service_name: Optional[str] = None,
) -> Callable:
    """Handler errors for sending report on email.

    Args:
        host(str, optional): = post of smtp server.
        port(int, optional): = port of smtp server.
        login(str, optional): = login for auth in smtp server.
        password(str, optional): password for auth in smtp server.
        from_addr(str, optional): the address sending this mail.
        to_addrs(str, list(str), optional): addresses to send this mail to.
        ssl(bool, optional): use SSL connection for smtp.
        service_name (optional): Service name.
    """

    def mailer_wrapper(
        *,
        exceptions: Optional[Union[Type[BaseException], Tuple[Type[BaseException], ...]]] = None,
        header: Optional[str] = None,
        as_attached: bool = False,
        datetime_format: str = "%Y-%m-%d %H:%M:%S",
        subject: Optional[str] = None,
        filename: Optional[str] = None,
    ):
        """Handler errors for sending report on email.

        Args:
            exceptions (exception, tuple(exception), optional): Exceptions for handle.
                Two and more - in tuple. Default - Exception.
            header (str, optional): first line in report message. Default -
            "Your program has crashed ☠️"
            as_attached (bool, optional): make report for sending as a file. Default - False.
            datetime_format (str, optional): format datetime for report.
            subject(str, optional): subject of the mail.
            filename(str, optional): filename for sending report as file.
                Default: datetime %Y-%m-%d %H_%M_%S.txt.
        """
        exceptions = exceptions or Exception
        params_for_send = {
            "host": host,
            "port": port,
            "login": login,
            "password": password,
            "from_addr": from_addr,
            "to_addrs": to_addrs,
            "ssl": ssl,
            "filename": filename,
            "subject": subject,
        }

        def decorator(func):
            func_name = func.__name__

            async def async_wrapper(*args, **kwargs):
                try:
                    return await func(*args, **kwargs)
                except exceptions as exc:
                    tback = traceback.format_exc()
                    report = await run_in_threadpool(
                        _report_maker,
                        tback=tback,
                        func_name=func_name,
                        header=header,
                        as_attached=as_attached,
                        service_name=service_name,
                        datetime_format=datetime_format,
                    )
                    await run_in_threadpool(
                        _report_mailer_handler,
                        report=report,
                        **params_for_send,
                    )
                    raise exc

            def wrapper(*args, **kwargs):
                try:
                    return func(*args, **kwargs)
                except exceptions as exc:
                    tback = traceback.format_exc()
                    report = _report_maker(
                        tback=tback,
                        func_name=func_name,
                        header=header,
                        as_attached=as_attached,
                        service_name=service_name,
                        datetime_format=datetime_format,
                    )
                    _report_mailer_handler(report=report, **params_for_send)
                    raise exc

            if asyncio.iscoroutinefunction(func) is True:
                return functools.wraps(func)(async_wrapper)
            return functools.wraps(func)(wrapper)

        return decorator

    return mailer_wrapper


def _report_maker(
    *,
    tback: str,
    func_name: Optional[str] = None,
    header: Optional[str] = None,
    as_attached: bool = False,
    service_name: Optional[str],
    datetime_format: Optional[str],
) -> Report:
    """
    Make report from
    Args:
        tback (str): traceback for report.
        func_name (str, optional): name of function when raised error.
        header (str, optional): first line in report message. Default -
        "Your program has crashed ☠️"
        as_attached (bool, optional): make report for sending as a file. Default - False.
    Returns:
        isinstance of Report obj.
    """
    return Report(tback, func_name, header, as_attached, service_name, datetime_format)



# pylint: disable=too-few-public-methods, too-many-instance-attributes
from datetime import datetime
from socket import gethostname
from typing import Optional


class Report:
    """Object for create report"""

    def __init__(
        self,
        tback: str,
        func_name: Optional[str],
        header: Optional[str],
        as_attached: bool,
        service_name: Optional[str],
        datetime_format: Optional[str],
    ):
        self._tback = tback
        self._func_name = func_name
        self._header = header
        self._host_name = gethostname()
        self._service_name = service_name
        self._as_attached = as_attached
        self._datetime_format = datetime_format
        self.report = None
        self.attach = None
        if self._as_attached is True:
            self._make_attach_report()
        else:
            self._make_text_report()

    def _make_text_report(self):
        """Formatting text report before sending."""
        crash_time = datetime.now().replace(microsecond=0)
        report = [
            "Your program has crashed ☠️",
            "Machine name: %s" % self._host_name,
            "Crash date: %s" % crash_time.strftime(self._datetime_format),
            "Traceback:",
            "%s" % self._tback,
        ]
        if self._header is not None:
            report[0] = "%s" % self._header
        if self._service_name is not None:
            report.insert(1, "Service: %s" % self._service_name)
        if self._func_name is not None:
            report.insert(3, "Main call: %s" % self._func_name)
        self.report = "\n".join(report)

    def _make_attach_report(self):
        """Formatting report with attach before sending."""
        crash_time = datetime.now().replace(microsecond=0)
        report = [
            "Your program has crashed ☠️",
            "Machine name: %s" % self._host_name,
            "Crash date: %s" % crash_time.strftime(self._datetime_format),
        ]
        if self._header is not None:
            report.insert(0, "%s" % self._header)
        if self._service_name is not None:
            report.insert(1, "Service: %s" % self._service_name)
        if self._func_name is not None:
            report.insert(2, "Main call: %s" % self._func_name)
        self.report = "\n".join(report)
        self.attach = self._tback



import os
import re
from itertools import chain

from setuptools import find_packages, setup


NAME_PACKAGE = "easy_notifyer"
DESCRIPTION = (
    "Easy bug reporter for small projects. "
    "Zero dependencies - download and run. Asyncio support."
)
EXTRAS = {
    "dev": [
        "pytest",
        "pytest-mock",
        "pytest-asyncio",
        "pylint",
        "flake8",
        "isort",
        "bump2version",
        "black",
    ]
}


def read(filename: str):
    """Open file"""
    return open(os.path.join(os.path.dirname(__file__), filename)).read().strip()


def read_version() -> str:
    """Read version from __init__.py"""
    regexp = re.compile(r'^__version__\W*=\W*"([\d.abrc]+)"')
    init_py = os.path.join(os.path.dirname(__file__), NAME_PACKAGE, "__init__.py")
    with open(init_py) as file:
        for line in file:
            match = regexp.match(line)
            if match is not None:
                return match.group(1)
        raise RuntimeError(f"Cannot find version in {NAME_PACKAGE}/__init__.py")


setup(
    name=NAME_PACKAGE,
    version=read_version(),
    description=DESCRIPTION,
    author="strpc",
    url="https://github.com/strpc/easy_notifyer",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    download_url="https://pypi.python.org/pypi/easy-notifyer",
    packages=find_packages(
        exclude=(
            "tests",
            "tests.*",
            "examples.*",
            "docs",
        )
    ),
    extras_require={
        "dev": EXTRAS["dev"],
        "all": list(chain.from_iterable(EXTRAS.values())),
    },
    python_requires=">=3.7, <4",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    license="MIT",
    keywords=["easy-notifyer", "telegram", "mailer", "mail-client"],
)



# pylint: disable=too-few-public-methods
import logging
import uuid
from io import BytesIO
from typing import BinaryIO, Dict, List, Optional, Tuple, Union
from urllib.error import HTTPError

from easy_notifyer.clients.requests import AsyncRequests, Requests
from easy_notifyer.utils import run_in_threadpool


logger = logging.getLogger(__name__)


class TelegramBase:
    """Base class of telegram"""

    def __init__(
        self,
        *,
        token: str,
        chat_id: Union[List[int], int, List[str], str],
        api_url: Optional[str] = None,
    ) -> None:
        """
        Args:
            token (str): Telegram bot token. To receive: https://core.telegram.org/bots#6-botfather.
            chat_id (int, str, list): Chat ids for send message.
        """
        self._token = token
        self._chat_ids = [chat_id] if isinstance(chat_id, (int, str)) else chat_id

        api_url = api_url or "https://api.telegram.org"
        api_base_url = api_url[:-1] if api_url.endswith("/") else api_url
        self._base_api_url = f"{api_base_url}/bot{self._token}/"

    @staticmethod
    def _prepare_attach(
        *,
        attach: Union[bytes, str, BinaryIO, Tuple[str, Union[BinaryIO, bytes]]],
        filename: Optional[str] = None,
    ) -> Dict:
        """Preparation of attach for sending to telegram

        Args:
            attach (bytes, str, binaryio, tuple): attach to send
            filename (str, optional): filename

        Returns:
            Dict: {'document': (filename, b'file)}
        """
        filename = filename or uuid.uuid4().hex
        files = {}
        if isinstance(attach, tuple) is True:
            files["document"] = attach
        elif isinstance(attach, bytes) is True:
            files["document"] = (filename, BytesIO(attach))
        elif hasattr(attach, "read") and isinstance(attach.read(0), bytes):
            files["document"] = (filename, attach)
        else:
            files["document"] = (filename, BytesIO(attach.encode()))
        return files


class TelegramAsync(TelegramBase):
    """Async client for telegram"""

    def __init__(
        self,
        *,
        token: str,
        chat_id: Union[List[int], int, List[str], str],
        api_url: Optional[str] = None,
    ) -> None:
        """
        Args:
            token (str): Telegram bot token. To receive: https://core.telegram.org/bots#6-botfather.
            chat_id (int, str, list): Chat ids for send message.
            api_url (str, optional): telegram api url.
        """
        super().__init__(token=token, chat_id=chat_id, api_url=api_url)
        self._client = AsyncRequests()

    async def _send_post(
        self,
        *,
        method_api: str,
        headers: Optional[Dict] = None,
        params: Optional[Dict] = None,
        body: Optional[Dict] = None,
        files: Optional[Dict] = None,
    ) -> None:
        """Send async post request.

        Args:
            method_api (str): method of api telegram
            headers (dict, optional): headers for of request.
            params (dict, optional): params of request.
            body (dict, optional): body of request.
            files (dict, optional): files of request in format ('filename.txt', b'filedata').
        """
        try:
            await self._client.post(
                url=self._base_api_url + method_api,
                headers=headers,
                params=params,
                body=body,
                files=files,
            )
        except HTTPError as error:
            logger.exception("Error. %s", error)
        except Exception:  # noqa
            logger.error("Send message to telegram error.")

    async def send_message(
        self,
        msg: str,
        disable_notification: bool = False,
        disable_web_page_preview: bool = False,
    ) -> None:
        """Send message.

        Args:
            msg (str): text of message.
            disable_notification (bool): True to disable notification of message.
            disable_web_page_preview (bool): True to disable web preview for links.
        """
        method_api = "sendMessage"

        for chat_id in self._chat_ids:
            body = {
                "chat_id": chat_id,
                "text": msg,
            }

            if disable_web_page_preview is True:
                body["disable_web_page_preview"] = True
            if disable_notification is True:
                body["disable_notification"] = True
            await self._send_post(method_api=method_api, body=body)

    async def send_attach(
        self,
        attach: Union[bytes, str, BinaryIO, Tuple[str, Union[BinaryIO, bytes]]],
        *,
        msg: Optional[str] = None,
        filename: Optional[str] = None,
        disable_notification: bool = False,
    ) -> None:
        """Send file.

        Args:
            attach (bytes, str, tuple): file to send. if tuple, then
                ('filename.txt', b'text of file').
            msg (str, optional): text of message.
            filename (str, optional): filename if attach is string or bytes.
            disable_notification (bool): True to disable notification of message.
        """
        method_api = "sendDocument"
        files = await run_in_threadpool(self._prepare_attach, attach=attach, filename=filename)

        params = {}
        if msg is not None:
            params["caption"] = msg
        if disable_notification is True:
            params["disable_notification"] = True

        for chat_id in self._chat_ids:
            params["chat_id"] = chat_id
            await self._send_post(method_api=method_api, params=params, files=files)


class Telegram(TelegramBase):
    """Client of telegram"""

    def __init__(
        self,
        *,
        token: str,
        chat_id: Union[List[int], int, List[str], str],
        api_url: Optional[str] = None,
    ) -> None:
        """
        Args:
            token (str): Telegram bot token. To receive: https://core.telegram.org/bots#6-botfather.
            chat_id (int, str, list): Chat ids for send message.
            api_url (str, optional): telegram api url.
        """
        super().__init__(token=token, chat_id=chat_id, api_url=api_url)
        self._client = Requests()

    def _send_post(
        self,
        *,
        method_api: str,
        headers: Optional[Dict] = None,
        params: Optional[Dict] = None,
        body: Optional[Dict] = None,
        files: Optional[Dict] = None,
    ) -> None:
        """Send post request.

        Args:
            method_api (str): method of api telegram
            headers (dict, optional): headers for of request.
            params (dict, optional): params of request.
            body (dict, optional): body of request.
            files (dict, optional): files of request in format ('filename.txt', b'filedata').
        """
        try:
            self._client.post(
                url=self._base_api_url + method_api,
                headers=headers,
                params=params,
                body=body,
                files=files,
            )
        except HTTPError as error:
            logger.exception("Error. %s", error)
        except Exception:  # noqa
            logger.error("Send message to telegram error.")

    def send_message(
        self,
        msg: str,
        disable_notification: bool = False,
        disable_web_page_preview: bool = False,
    ) -> None:
        """Send message.

        Args:
            msg (str): text of message.
            disable_notification (bool): True to disable notification of message.
            disable_web_page_preview (bool): True to disable web preview for links.
        """
        method_api = "sendMessage"

        for chat_id in self._chat_ids:
            body = {
                "chat_id": chat_id,
                "text": msg,
            }
            if disable_web_page_preview is True:
                body["disable_web_page_preview"] = True
            if disable_notification is True:
                body["disable_notification"] = True
            self._send_post(method_api=method_api, body=body)

    def send_attach(
        self,
        attach: Union[bytes, str, BinaryIO, Tuple[str, Union[BinaryIO, bytes]]],
        *,
        msg: Optional[str] = None,
        filename: Optional[str] = None,
        disable_notification: bool = False,
    ) -> None:
        """Send file.

        Args:
            attach (bytes, str, tuple): file to send. if tuple, then
                ('filename.txt', b'text of file').
            msg (str, optional): text of message.
            filename (str, optional): filename if attach is string or bytes.
            disable_notification (bool): True to disable notification of message.
        """
        method_api = "sendDocument"
        files = self._prepare_attach(attach=attach, filename=filename)

        params = {}
        if msg is not None:
            params["caption"] = msg
        if disable_notification is True:
            params["disable_notification"] = True

        for chat_id in self._chat_ids:
            params["chat_id"] = chat_id
            self._send_post(method_api=method_api, params=params, files=files)
